#include <config.h>
module quark_wilson_class
!
!= Definition of Wilson quark related parameters and routines
!
!== version
!
! $Id: quark_wilson_class.F90,v 1.20 2011/09/25 06:13:16 ishikawa Exp $
!
  use lattice_class
  use hmc_status_class
  use field_gauge_class
  use field_fermion_class
  use action_base_class
  use comlib
  implicit none
  private
  public :: quark_wilson
  public :: new
  public :: delete
  public :: assign
  public :: read
  public :: save_config
  public :: print
  public :: set_gaussian_noise
  public :: assign_mult_hop
  public :: assign_mult_eoprec_wd
  public :: assign_mult_eoprec_wd_dag
  public :: assign_mult_Qee
  public :: assign_mult_Oee
  public :: set_id
  public :: get_id
  public :: get_kappa
  public :: get_nflavor
  public :: set_kappa
  public :: set_nflavor

  character(len=*), parameter :: ACTION_NAME = "Wilson Quark"

  type, extends(action_parameters) :: quark_wilson
  !
  ! Wilson quark type extended from the above type (extends from action_parameters)
  !
    integer,  private :: nflavor    ! number of flavor of this quark, 1 or 2
    real(DP), private :: kappa      ! hopping parameter
  end type

  interface new
    module procedure new_quark_wilson
  end interface

  interface delete
    module procedure delete_quark_wilson
  end interface

  interface assign
    module procedure assign_quark_wilson
  end interface

  interface read
  !
  ! read action parameters from formatted file
  !
    module procedure read_quark_wilson
  end interface

  interface save_config
  !
  ! save action parameters to configuraion file as unformatted data
  !
    module procedure save_config_quark_wilson
  end interface

  interface print
    module procedure print_quark_wilson
  end interface

  interface assign_mult_hop
  !
  ! Hopping matrix multiplication
  !
    module procedure  assign_mult_hopping_eo
    module procedure  assign_mult_hopping
  end interface

  interface assign_mult_eoprec_wd
  !
  ! Even/odd site preconditioned Wilson-Dirac matrix/operator multiplication
  !
    module procedure  assign_mult_eoprec_wd_eo
  end interface

  interface assign_mult_eoprec_wd_dag
  !
  ! Hermitian conjugate Even/odd site preconditioned Wilson-Dirac
  !  matrix/operator multiplication
  !
    module procedure  assign_mult_eoprec_wd_eo_dag
  end interface

  interface assign_mult_Qee
  !
  ! Hermitian quared Even/odd site preconditioned Wilson-Dirac
  ! matrix/operator multiplication
  !
  !  Qee = Dee^+ Dee
  !
    module procedure  assign_mult_DdagD_ee_wq
  end interface

  interface assign_mult_Oee
  !
  ! Hermitian quared Even/odd site preconditioned Wilson-Dirac
  ! matrix/operator multiplication
  !
  !  Oee = Dee Dee^+
  !
    module procedure  assign_mult_DDdag_ee_wq
  end interface

  interface get_kappa
  !
  ! get quark kappa
  !
    module procedure get_kappa
  end interface

  interface get_nflavor
  !
  ! get number of flavor
  !
    module procedure get_nflavor
  end interface

  interface set_kappa
  !
  ! set quark kappa
  !
    module procedure set_kappa
  end interface

  interface set_nflavor
  !
  ! set number of flavor
  !
    module procedure set_nflavor
  end interface
contains

function get_kappa(this) result(kappa)
  type(quark_wilson), intent(in) :: this
  real(DP) :: kappa
  kappa = this%kappa
  return
end function

function get_nflavor(this) result(nf)
  type(quark_wilson), intent(in) :: this
  integer :: nf
  nf = this%nflavor
  return
end function

subroutine set_kappa(this,kappa)
  type(quark_wilson), intent(inout) :: this
  real(DP),           intent(in)    :: kappa
  this%kappa = kappa
  return
end subroutine

subroutine set_nflavor(this,nflavor)
  type(quark_wilson), intent(inout) :: this
  integer,            intent(in)    :: nflavor
  this%nflavor = nflavor
  return
end subroutine

subroutine new_quark_wilson(this,id)
  implicit none
  type(quark_wilson), intent(inout) :: this
  integer,            intent(in)    :: id
  this%kappa   = 0.0_DP
  this%nflavor = 0
  this%action_name = ACTION_NAME
  return
end subroutine

subroutine delete_quark_wilson(this)
  implicit none
  type(quark_wilson), intent(inout) :: this
  this%action_name = REPEAT(' ',LEN(this%action_name))
  return
end subroutine

subroutine assign_quark_wilson(qp1,qp2)
  implicit none
  type(quark_wilson), intent(inout) :: qp1
  type(quark_wilson), intent(in)    :: qp2
  qp1 = qp2
  return
end subroutine

subroutine read_quark_wilson(this,iout)
  implicit none
  type(quark_wilson), intent(inout) :: this
  integer, intent(in) :: iout
  if (nodeid==0) then
    read(iout,*)this%nflavor
    read(iout,*)this%kappa
  endif
#ifndef _singlePU
  call comlib_bcast(this%kappa,0)
  call comlib_bcast(this%nflavor,0)
#endif
  return
end subroutine

subroutine save_config_quark_wilson(this,iout)
  implicit none
  type(quark_wilson), intent(in) :: this
  integer, intent(in) :: iout
  write(iout)this%nflavor
  write(iout)this%kappa
  return
end subroutine

subroutine print_quark_wilson(this)
  implicit none
  type(quark_wilson), intent(inout) :: this
  write(*,'(14X,"    Num of Flavor :",I3)')   this%nflavor
  write(*,'(14X,"            Kappa :",F12.8)')this%kappa
  return
end subroutine

subroutine assign_mult_hopping_eo(this,Mye,yo,u)
!
!= Multiply Hopping operator
!
!   Mye = Meo[u] yo (hoppng odd->even/even->odd)
!
  use field_gauge_class
  use counter_class
  implicit none
  type(quark_wilson),      intent(in)    :: this ! Wilson quark parameters
  type(field_quark_eo_wg), intent(out)   :: Mye  ! even/odd site fermion field
  type(field_quark_eo_wg), intent(inout) ::  yo  ! odd/even site fermion field
  type(vfield_gluon_wg),   intent(in)    ::   u  ! gauge field

  call copy_boundary(yo)
  call mult_hopping_tzyx_eo(Mye,yo,u)
  call inc(mult_iter)

  return
end subroutine

subroutine assign_mult_hopping(this,My,y,u)
!
!= Multiply Hopping operator
!
!   My = M[u] y
!
  use field_gauge_class
  implicit none
  type(quark_wilson),    intent(in)    :: this ! Wilson quark parameters
  type(field_quark_wg),  intent(out)   ::  My  ! even/odd site fermion field
  type(field_quark_wg),  intent(inout) ::   y  ! odd/even site fermion field
  type(vfield_gluon_wg), intent(in)    ::   u  ! gauge field
  call assign_mult_hop(this,My%eo(0),y%eo(1),u)
  call assign_mult_hop(this,My%eo(1),y%eo(0),u)
  return
end subroutine

subroutine assign_mult_eoprec_wd_eo(this,yde,ye,u)
!
!= Multiply even/odd-site preconditioned Dirac operator :
!        v
!  yde = Dee ye,
!
!  where  v
!         Dee = 1ee - kappa^2 Meo Moe
! 
  implicit none
  type(quark_wilson),      intent(in)    :: this  ! wilson quark parameter
  type(field_quark_eo_wg), intent(inout) :: yde   ! even/odd fermion field
  type(field_quark_eo_wg), intent(inout) ::  ye   ! even/odd fermion field / boundary copied
  type(vfield_gluon_wg),   intent(in)    ::   u   ! gauge field

  type(field_quark_eo_wg), allocatable :: Myo
  real(DP) :: kappa2

  allocate(Myo)
  kappa2=-(this%kappa)**2

!***********************************************************
!* Calc:
!*        v
!*  yde = D ye
!*      = ( 1 - kappa**2 Meo Moe ) ye
!*      
!***********************************************************

!***********************************
!*  Myo = F^-1oo Moe ye
!*  yde = F^-1ee Meo Myo = F^-1ee Meo F^-1oo Moe ye
!***********************************
  call assign_mult_hop(this,Myo, ye,u)
  call assign_mult_hop(this,yde,Myo,u)

!***********************************
!*  yde = ye - kappa^2 yde
!*      = ye - kappa^2 Meo Moe ye
!***********************************
  call accum_mult_add(yde,kappa2,ye)

  deallocate(Myo)

  return
end subroutine

subroutine assign_mult_eoprec_wd_eo_dag(phys_param,dye,ye,u)
!
!= Multiply Hermitian conjugate of the even/odd-site preconditioned wilson-dirac operator
!
!        v
!  dye = Dee^+ ye
!
!  v
!  Dee^+ = 1ee - gamma_5 Meo Foo \ Moe Fee \ gamma_5
!
  implicit none
  type(quark_wilson), intent(in) :: phys_param
  type(field_quark_eo_wg), intent(inout) :: dye
  type(field_quark_eo_wg), intent(in)    ::  ye
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(field_quark_eo_wg), allocatable :: wo
  real(DP) :: kappa,kappa2

  kappa = get_kappa(phys_param)

  !=============================
  ! dye = Fee \ gamma_5 ye
  !=============================
  call assign_mult_gamma5(dye,ye)
  
  !=============================
  ! wo = Moe dye
  !=============================
  allocate(wo)
  call assign_mult_hop(phys_param,wo,dye,u)

  !=============================
  ! dye = gamma_5 Meo wo
  !=============================
  call assign_mult_hop(phys_param,dye,wo,u)
  call accum_mult_gamma5(dye)
  deallocate(wo)

  !=============================
  ! dye = -kappa^2 dye + ye
  !=============================
  kappa2 = -kappa**2
  call accum_mult_add(dye,kappa2,ye)

  return
end subroutine

subroutine assign_mult_DdagD_ee_wq(phys_param,qye,ye,u)
!
!= Multiply Hermitian squared the even/odd-site preconditioned wilson-dirac
!  operator
!
!        v     v        v
!  qye = Dee^+ Dee ye = Qee ye
!
  type(quark_wilson), intent(in) :: phys_param
  type(field_quark_eo_wg), intent(inout) :: qye
  type(field_quark_eo_wg), intent(inout) ::  ye
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(field_quark_eo_wg), allocatable   ::  we
  allocate(we)
  call assign_mult_eoprec_wd(    phys_param, we,ye,u)
  call assign_mult_eoprec_wd_dag(phys_param,qye,we,u)
  deallocate(we)
  return
end subroutine

subroutine assign_mult_DDdag_ee_wq(phys_param,qye,ye,u)
!
!= Multiply Hermitian squared the even/odd-site preconditioned wilson-dirac
!  operator
!
!        v   v          v
!  qye = Dee Dee^+ ye = Oee ye
!
  type(quark_wilson), intent(in) :: phys_param
  type(field_quark_eo_wg), intent(inout) :: qye
  type(field_quark_eo_wg), intent(inout) ::  ye
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(field_quark_eo_wg), allocatable   ::  we
  allocate(we)
  call assign_mult_eoprec_wd_dag(phys_param, we,ye,u)
  call assign_mult_eoprec_wd(    phys_param,qye,we,u)
  deallocate(we)
  return
end subroutine

end module
